@@ -77,7 +77,6 @@ static char * connectable_interfaces[] = {
 #define BLUEZ_AUDIOSINK_INTERFACE (detectable_interfaces[1])
 
 static DBusGConnection *connection = NULL;
-static BluetoothClient *bluetooth_client = NULL;
 
 #define BLUETOOTH_CLIENT_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), \
 				BLUETOOTH_TYPE_CLIENT, BluetoothClientPrivate))
@@ -408,35 +407,47 @@ device_list_nodes (DBusGProxy *device, BluetoothClient *client, gboolean connect
 /* Short names from Table 2 at:
  * https://www.bluetooth.org/Technical/AssignedNumbers/service_discovery.htm */
 static const char *
-uuid16_to_string (guint uuid16)
+uuid16_to_string (guint uuid16, const char *uuid)
 {
 	switch (uuid16) {
+	case 0x1101:
+		return "SerialPort";
 	case 0x1103:
 		return "DialupNetworking";
+	case 0x1104:
+		return "IrMCSync";
 	case 0x1105:
 		return "OBEXObjectPush";
 	case 0x1106:
 		return "OBEXFileTransfer";
+	case 0x110A:
+		return "AudioSource";
 	case 0x110B:
 		return "AudioSink";
 	case 0x110c:
 		return "A/V_RemoteControlTarget";
 	case 0x110e:
 		return "A/V_RemoteControl";
+	case 0x1112:
+		return "Headset_-_AG";
 	case 0x1115:
 		return "PANU";
 	case 0x1116:
 		return "NAP";
 	case 0x1117:
 		return "GN";
+	case 0x111F:
+		return "HandsfreeAudioGateway";
 	case 0x1124:
 		return "HumanInterfaceDeviceService";
+	case 0x112F:
+		return "Phonebook_Access_-_PSE";
 	case 0x1000: /* ServiceDiscoveryServerServiceClassID */
 	case 0x1200: /* PnPInformation */
 		/* Those are ignored */
 		return NULL;
 	default:
-		g_debug ("Unhandled UUID 0x%x", uuid16);
+		g_debug ("Unhandled UUID %s (0x%x)", uuid, uuid16);
 		return NULL;
 	}
 }
@@ -458,7 +469,7 @@ bluetooth_uuid_to_string (const char *uuid)
 	if (uuid16 == 0)
 		return NULL;
 
-	return uuid16_to_string (uuid16);
+	return uuid16_to_string (uuid16, uuid);
 }
 
 static char **
@@ -1109,11 +1120,14 @@ static void bluetooth_client_class_init(BluetoothClientClass *klass)
 
 BluetoothClient *bluetooth_client_new(void)
 {
+	static BluetoothClient *bluetooth_client = NULL;
+
 	if (bluetooth_client != NULL)
 		return g_object_ref(bluetooth_client);
 
-	bluetooth_client = BLUETOOTH_CLIENT(g_object_new(BLUETOOTH_TYPE_CLIENT,
-									NULL));
+	bluetooth_client = BLUETOOTH_CLIENT (g_object_new (BLUETOOTH_TYPE_CLIENT, NULL));
+	g_object_add_weak_pointer (G_OBJECT (bluetooth_client),
+				   (gpointer) &bluetooth_client);
 
 	DBG("client %p", bluetooth_client);
 
@@ -1326,6 +1340,7 @@ gboolean bluetooth_client_stop_discovery(BluetoothClient *client)
 typedef struct {
 	BluetoothClientCreateDeviceFunc func;
 	gpointer data;
+	BluetoothClient *client;
 } CreateDeviceData;
 
 static void create_device_callback(DBusGProxy *proxy,
@@ -1344,8 +1359,9 @@ static void create_device_callback(DBusGProxy *proxy,
 	}
 
 	if (devdata->func)
-		devdata->func(bluetooth_client, path, devdata->data);
+		devdata->func(devdata->client, path, devdata->data);
 
+	g_object_unref (devdata->client);
 	g_object_unref(proxy);
 }
 
@@ -1372,6 +1388,7 @@ gboolean bluetooth_client_create_device(BluetoothClient *client,
 
 	devdata->func = func;
 	devdata->data = data;
+	devdata->client = g_object_ref (client);
 
 	if (agent != NULL)
 		call = dbus_g_proxy_begin_call_with_timeout(adapter,
@@ -1379,7 +1396,7 @@ gboolean bluetooth_client_create_device(BluetoothClient *client,
 				devdata, g_free, 90 * 1000,
 				G_TYPE_STRING, address,
 				DBUS_TYPE_G_OBJECT_PATH, agent,
-				G_TYPE_STRING, "DisplayOnly", G_TYPE_INVALID);
+				G_TYPE_STRING, "DisplayYesNo", G_TYPE_INVALID);
 	else
 		call = dbus_g_proxy_begin_call_with_timeout(adapter,
 				"CreateDevice", create_device_callback,
@@ -1424,6 +1441,7 @@ gboolean bluetooth_client_set_trusted(BluetoothClient *client,
 typedef struct {
 	BluetoothClientConnectFunc func;
 	gpointer data;
+	BluetoothClient *client;
 	/* used for disconnect */
 	GList *services;
 } ConnectData;
@@ -1443,8 +1461,9 @@ static void connect_callback(DBusGProxy *proxy,
 	}
 
 	if (conndata->func)
-		conndata->func(bluetooth_client, retval, conndata->data);
+		conndata->func(conndata->client, retval, conndata->data);
 
+	g_object_unref (conndata->client);
 	g_object_unref(proxy);
 }
 
@@ -1501,6 +1520,7 @@ gboolean bluetooth_client_connect_service(BluetoothClient *client,
 
 	conndata->func = func;
 	conndata->data = data;
+	conndata->client = g_object_ref (client);
 
 	dbus_g_proxy_begin_call(proxy, "Connect",
 				connect_callback, conndata, g_free,
@@ -1535,8 +1555,9 @@ static void disconnect_callback(DBusGProxy *proxy,
 	}
 
 	if (conndata->func)
-		conndata->func(bluetooth_client, TRUE, conndata->data);
+		conndata->func(conndata->client, TRUE, conndata->data);
 
+	g_object_unref (conndata->client);
 	g_object_unref(proxy);
 	g_free (conndata);
 }
@@ -1606,6 +1627,7 @@ gboolean bluetooth_client_disconnect_service (BluetoothClient *client,
 
 	conndata->func = func;
 	conndata->data = data;
+	conndata->client = g_object_ref (client);
 
 	if (table == NULL) {
 		dbus_g_proxy_begin_call(proxy, "Disconnect",
