@@ -32,18 +32,19 @@
 #include <dbus/dbus-glib.h>
 
 #include <bluetooth-client.h>
-#include <helper.h>
+#include <bluetooth-killswitch.h>
+#include <bluetooth-chooser.h>
+#include <bluetooth-plugin-manager.h>
 
 #include "adapter.h"
 #include "general.h"
-#include "bluetooth-killswitch.h"
 
 static BluetoothClient *client;
 static GtkTreeModel *adapter_model;
+static BluetoothKillswitch *killswitch;
 
-static GtkWidget *killswitch_page;
-static BluetoothKillswitch *killswitch = NULL;
-static gboolean has_killswitch_page = FALSE;
+#define KILLSWITCH_PAGE_NUM(n) (gtk_notebook_get_n_pages (n) - 2)
+#define NO_ADAPTERS_PAGE_NUM(n) (-1)
 
 typedef struct adapter_data adapter_data;
 struct adapter_data {
@@ -51,73 +52,32 @@ struct adapter_data {
 	GtkWidget *notebook;
 	GtkWidget *child;
 	GtkWidget *button_discoverable;
-	GtkWidget *button_powered;
 	GtkWidget *name_vbox, *devices_table;
 	GtkWidget *entry;
 	GtkWidget *button_delete;
-	GtkWidget *button_trusted;
 	GtkWidget *button_disconnect;
-	GtkTreeSelection *selection;
+	GtkWidget *chooser;
 	GtkTreeRowReference *reference;
 	guint signal_discoverable;
-	guint signal_powered;
 	gboolean powered;
 	gboolean discoverable;
+	gboolean is_default;
 	guint timeout_value;
 	int name_changed;
-	gboolean bring_forward;
 };
 
 static void update_visibility(adapter_data *adapter);
-static GtkWidget * create_killswitch_page (void);
 
 static void block_signals(adapter_data *adapter)
 {
 	g_signal_handler_block(adapter->button_discoverable,
 						adapter->signal_discoverable);
-	g_signal_handler_block(adapter->button_powered,
-						adapter->signal_powered);
 }
 
 static void unblock_signals(adapter_data *adapter)
 {
 	g_signal_handler_unblock(adapter->button_discoverable,
 						adapter->signal_discoverable);
-	g_signal_handler_unblock(adapter->button_powered,
-						adapter->signal_powered);
-}
-
-static void update_tab_label(GtkNotebook *notebook,
-					GtkWidget *child, const char *name)
-{
-	GtkWidget *label;
-
-	gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), child,
-				name && name[0] != '\0' ? name : _("Unnamed Adapter"));
-
-	label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), child);
-	gtk_label_set_max_width_chars(GTK_LABEL(label), 20);
-	gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
-}
-
-static void powered_changed_cb(GtkWidget *button, gpointer user_data)
-{
-	adapter_data *adapter = user_data;
-	GValue powered = { 0 };
-
-	adapter->powered = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button));
-	/* Cheat and update the visibility straight away */
-	update_visibility (adapter);
-
-	g_value_init(&powered, G_TYPE_BOOLEAN);
-	g_value_set_boolean(&powered, adapter->powered);
-
-	dbus_g_proxy_call(adapter->proxy, "SetProperty", NULL,
-					G_TYPE_STRING, "Powered",
-					G_TYPE_VALUE, &powered,
-					G_TYPE_INVALID, G_TYPE_INVALID);
-
-	g_value_unset(&powered);
 }
 
 static void discoverable_changed_cb(GtkWidget *button, gpointer user_data)
@@ -181,64 +141,23 @@ static gboolean focus_callback(GtkWidget *editable,
 	return FALSE;
 }
 
-static void update_buttons(adapter_data *adapter, gboolean bonded,
-					gboolean trusted, gboolean connected)
+static void update_buttons(adapter_data *adapter, gboolean selected, gboolean connected)
 {
-	if (trusted) {
-		gtk_button_set_label(GTK_BUTTON(adapter->button_trusted), _("Distrust"));
-	} else {
-		gtk_button_set_label(GTK_BUTTON(adapter->button_trusted), _("Trust"));
-	}
+	gtk_widget_set_sensitive(adapter->button_delete, selected);
 	gtk_widget_set_sensitive(adapter->button_disconnect, connected);
 }
 
-static void select_callback(GtkTreeSelection *selection, gpointer user_data)
+static void
+device_selected_cb(GObject *object,
+		   GParamSpec *spec,
+		   adapter_data *adapter)
 {
-	adapter_data *adapter = user_data;
-	DBusGProxy *proxy;
-	GtkTreeModel *model;
-	GtkTreeIter iter;
-	gboolean selected;
-	gboolean paired = FALSE, trusted = FALSE, connected = FALSE;
+	char *address;
+	gboolean connected;
 
-	selected = gtk_tree_selection_get_selected(selection, &model, &iter);
-
-	if (selected == TRUE) {
-		gtk_tree_model_get(model, &iter,
-				BLUETOOTH_COLUMN_PROXY, &proxy,
-				BLUETOOTH_COLUMN_PAIRED, &paired,
-				BLUETOOTH_COLUMN_TRUSTED, &trusted,
-				BLUETOOTH_COLUMN_CONNECTED, &connected, -1);
-
-		if (proxy != NULL) {
-			paired = TRUE;
-			g_object_unref(proxy);
-		}
-	}
-
-	update_buttons(adapter, paired, trusted, connected);
-
-	gtk_widget_set_sensitive(adapter->button_trusted, selected);
-	gtk_widget_set_sensitive(adapter->button_delete, selected);
-}
-
-static void row_callback(GtkTreeModel *model, GtkTreePath  *path,
-					GtkTreeIter *iter, gpointer user_data)
-{
-	adapter_data *adapter = user_data;
-	gboolean bonded = FALSE, trusted = FALSE, connected = FALSE;
-
-	if (gtk_tree_selection_iter_is_selected(adapter->selection,
-							iter) == FALSE)
-		return;
-
-	gtk_tree_model_get(model, iter, BLUETOOTH_COLUMN_PAIRED, &bonded,
-					BLUETOOTH_COLUMN_TRUSTED, &trusted,
-					BLUETOOTH_COLUMN_CONNECTED, &connected, -1);
-
-	update_buttons(adapter, bonded, trusted, connected);
-
-	gtk_widget_set_sensitive(adapter->button_delete, TRUE);
+	address = bluetooth_chooser_get_selected_device (BLUETOOTH_CHOOSER (adapter->chooser));
+	connected = bluetooth_chooser_get_selected_device_is_connected (BLUETOOTH_CHOOSER (adapter->chooser));
+	update_buttons(adapter, (address != NULL), connected);
 }
 
 static void wizard_callback(GtkWidget *button, gpointer user_data)
@@ -249,29 +168,24 @@ static void wizard_callback(GtkWidget *button, gpointer user_data)
 		g_printerr("Couldn't execute command: %s\n", command);
 }
 
-static gboolean show_confirm_dialog(void)
+static gboolean show_confirm_dialog(const char *name)
 {
 	GtkWidget *dialog;
 	gint response;
-	gchar *text;
 
-	text = g_strdup_printf("<big><b>%s</b></big>\n\n%s",
-				_("Remove from list of known devices?"),
-				_("If you delete the device, you have to "
-					"set it up again before next use."));
-	dialog = gtk_message_dialog_new_with_markup(NULL, GTK_DIALOG_MODAL,
-				GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE, NULL);
-	gtk_message_dialog_set_markup(GTK_MESSAGE_DIALOG(dialog), text);
-	g_free(text);
+	dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_MODAL,
+					 GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
+					 _("Remove '%s' from list of known devices?"), name);
+	g_object_set (G_OBJECT (dialog), "secondary-text",
+		      _("If you delete the device, you have to set it up again before next use."),
+		      NULL);
 
-	gtk_dialog_add_button(GTK_DIALOG(dialog), GTK_STOCK_CANCEL,
-							GTK_RESPONSE_CANCEL);
-	gtk_dialog_add_button(GTK_DIALOG(dialog), GTK_STOCK_DELETE,
-							GTK_RESPONSE_ACCEPT);
+	gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
+	gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_DELETE, GTK_RESPONSE_ACCEPT);
 
-	response = gtk_dialog_run(GTK_DIALOG(dialog));
+	response = gtk_dialog_run (GTK_DIALOG (dialog));
 
-	gtk_widget_destroy(dialog);
+	gtk_widget_destroy (dialog);
 
 	if (response == GTK_RESPONSE_ACCEPT)
 		return TRUE;
@@ -282,75 +196,61 @@ static gboolean show_confirm_dialog(void)
 static void delete_callback(GtkWidget *button, gpointer user_data)
 {
 	adapter_data *adapter = user_data;
-	GtkTreeModel *model;
-	GtkTreeIter iter;
+	GValue value = { 0, };
 	DBusGProxy *device;
 	const char *path;
+	char *address, *name;
 
-	if (gtk_tree_selection_get_selected(adapter->selection,
-						&model, &iter) == FALSE)
+	if (bluetooth_chooser_get_selected_device_info (BLUETOOTH_CHOOSER (adapter->chooser),
+							"proxy", &value) == FALSE) {
 		return;
-
-	gtk_tree_model_get(model, &iter, BLUETOOTH_COLUMN_PROXY, &device, -1);
+	}
+	address = bluetooth_chooser_get_selected_device (BLUETOOTH_CHOOSER (adapter->chooser));
+	name = bluetooth_chooser_get_selected_device_name (BLUETOOTH_CHOOSER (adapter->chooser));
+	if (address == NULL || name == NULL) {
+		g_value_unset (&value);
+		g_free (address);
+		g_free (name);
+		return;
+	}
+	device = g_value_dup_object (&value);
+	g_value_unset (&value);
 
 	if (device == NULL)
 		return;
 
 	path = dbus_g_proxy_get_path(device);
 
-	if (show_confirm_dialog() == TRUE) {
-		dbus_g_proxy_call(adapter->proxy, "RemoveDevice", NULL,
-					DBUS_TYPE_G_OBJECT_PATH, path,
-					G_TYPE_INVALID, G_TYPE_INVALID);
+	if (show_confirm_dialog (name) == TRUE) {
+		GError *err = NULL;
+		if (dbus_g_proxy_call (adapter->proxy, "RemoveDevice", &err,
+				       DBUS_TYPE_G_OBJECT_PATH, path,
+				       G_TYPE_INVALID, G_TYPE_INVALID) == FALSE) {
+			g_warning ("Failed to remove device %s: %s", address,
+				   err->message);
+			g_error_free (err);
+		} else {
+			bluetooth_plugin_manager_device_deleted (address);
+		}
 	}
 
-	g_object_unref(device);
-}
-
-static void trusted_callback(GtkWidget *button, gpointer user_data)
-{
-	adapter_data *adapter = user_data;
-	GValue value = { 0 };
-	GtkTreeModel *model;
-	GtkTreeIter iter;
-	DBusGProxy *device;
-	gboolean trusted = FALSE;
-
-	if (gtk_tree_selection_get_selected(adapter->selection,
-						&model, &iter) == FALSE)
-		return;
-
-	gtk_tree_model_get(model, &iter, BLUETOOTH_COLUMN_PROXY, &device,
-					BLUETOOTH_COLUMN_TRUSTED, &trusted, -1);
-
-	if (device == NULL)
-		return;
-
-	g_value_init(&value, G_TYPE_BOOLEAN);
-
-	g_value_set_boolean(&value, !trusted);
-
-	dbus_g_proxy_call(device, "SetProperty", NULL,
-			G_TYPE_STRING, "Trusted",
-			G_TYPE_VALUE, &value, G_TYPE_INVALID, G_TYPE_INVALID);
-
-	g_value_unset(&value);
-
+	g_free (address);
+	g_free (name);
 	g_object_unref(device);
 }
 
 static void disconnect_callback(GtkWidget *button, gpointer user_data)
 {
 	adapter_data *adapter = user_data;
-	GtkTreeModel *model;
-	GtkTreeIter iter;
+	GValue value = { 0, };
 	DBusGProxy *device;
 
-	if (gtk_tree_selection_get_selected(adapter->selection,
-						&model, &iter) == FALSE)
+	if (bluetooth_chooser_get_selected_device_info (BLUETOOTH_CHOOSER (adapter->chooser),
+							"proxy", &value) == FALSE) {
 		return;
-
-	gtk_tree_model_get(model, &iter, BLUETOOTH_COLUMN_PROXY, &device, -1);
+	}
+	device = g_value_dup_object (&value);
+	g_value_unset (&value);
 
 	if (device == NULL)
 		return;
@@ -363,43 +263,38 @@ static void disconnect_callback(GtkWidget *button, gpointer user_data)
 	gtk_widget_set_sensitive(button, FALSE);
 }
 
-static gboolean device_filter(GtkTreeModel *model,
-					GtkTreeIter *iter, gpointer user_data)
+static void
+set_current_page (GtkNotebook *notebook)
 {
-	DBusGProxy *proxy;
-	gboolean active;
-
-	gtk_tree_model_get(model, iter, BLUETOOTH_COLUMN_PROXY, &proxy,
-					BLUETOOTH_COLUMN_PAIRED, &active, -1);
-
-	if (proxy != NULL) {
-		active = TRUE;
-		g_object_unref(proxy);
+	if (gtk_tree_model_iter_n_children (adapter_model, NULL) == 0) {
+		if (bluetooth_killswitch_has_killswitches (killswitch) != FALSE) {
+			gtk_notebook_set_current_page (notebook,
+						       KILLSWITCH_PAGE_NUM(notebook));
+		} else {
+			gtk_notebook_set_current_page (notebook,
+						       NO_ADAPTERS_PAGE_NUM(notebook));
+		}
 	}
-
-	return active;
 }
 
 static void create_adapter(adapter_data *adapter)
 {
 	GHashTable *hash = NULL;
 	GValue *value;
+	DBusGProxy *default_proxy;
 	const gchar *address, *name;
 	gboolean powered, discoverable;
 	guint timeout;
 
 	GtkWidget *mainbox;
 	GtkWidget *vbox;
+	GtkWidget *alignment;
 	GtkWidget *table;
 	GtkWidget *label;
 	GtkWidget *image;
 	GtkWidget *button;
 	GtkWidget *entry;
 	GtkWidget *buttonbox;
-	GtkWidget *scrolled;
-	GtkWidget *tree;
-	GtkTreeModel *model;
-	GtkTreeSelection *selection;
 	int page_num;
 
 	dbus_g_proxy_call(adapter->proxy, "GetProperties", NULL, G_TYPE_INVALID,
@@ -434,27 +329,24 @@ static void create_adapter(adapter_data *adapter)
 	adapter->discoverable = discoverable;
 	adapter->timeout_value = timeout;
 
-	mainbox = gtk_vbox_new(FALSE, 18);
+	default_proxy = bluetooth_client_get_default_adapter (client);
+	if (default_proxy != NULL) {
+		adapter->is_default = g_str_equal (dbus_g_proxy_get_path (default_proxy),
+						   dbus_g_proxy_get_path (adapter->proxy));
+		g_object_unref (default_proxy);
+	}
+
+	mainbox = gtk_vbox_new(FALSE, 6);
 	gtk_container_set_border_width(GTK_CONTAINER(mainbox), 12);
 
 	page_num = gtk_notebook_prepend_page(GTK_NOTEBOOK(adapter->notebook),
 							mainbox, NULL);
-	update_tab_label(GTK_NOTEBOOK(adapter->notebook), mainbox, name);
 
 	adapter->child = mainbox;
 
 	vbox = gtk_vbox_new(FALSE, 6);
 	gtk_box_pack_start(GTK_BOX(mainbox), vbox, FALSE, TRUE, 0);
 
-	/* The powered checkbox */
-	button = gtk_check_button_new_with_mnemonic (_("_Powered"));
-	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), powered);
-
-	adapter->button_powered = button;
-	adapter->signal_powered = g_signal_connect(G_OBJECT(button), "toggled",
-						   G_CALLBACK(powered_changed_cb), adapter);
-	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
-
 	/* The discoverable checkbox */
 	button = gtk_check_button_new_with_mnemonic (_("_Discoverable"));
 	if (powered == FALSE)
@@ -482,10 +374,15 @@ static void create_adapter(adapter_data *adapter)
 	label = create_label(_("Friendly name"));
 	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
 
+	alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
+	gtk_widget_show (alignment);
+	gtk_box_pack_start (GTK_BOX (vbox), alignment, TRUE, TRUE, 0);
+	gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), 0, 0, 12, 0);
+
 	entry = gtk_entry_new();
 	gtk_entry_set_max_length(GTK_ENTRY(entry), 248);
 	gtk_widget_set_size_request(entry, 240, -1);
-	gtk_box_pack_start(GTK_BOX(vbox), entry, TRUE, TRUE, 0);
+	gtk_container_add (GTK_CONTAINER (alignment), entry);
 
 	if (name != NULL)
 		gtk_entry_set_text(GTK_ENTRY(entry), name);
@@ -508,32 +405,26 @@ static void create_adapter(adapter_data *adapter)
 	gtk_table_attach(GTK_TABLE(table), label, 0, 2, 0, 1,
 			 GTK_EXPAND | GTK_FILL, GTK_SHRINK, 0, 6);
 
-	scrolled = gtk_scrolled_window_new(NULL, NULL);
-	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
-				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
-	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled),
-							GTK_SHADOW_OUT);
-	gtk_table_attach(GTK_TABLE(table), scrolled, 0, 1, 1, 2,
-			 GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 6, 6);
+	/* Note that this will only ever show the devices on the default
+	 * adapter, this is on purpose */
+	adapter->chooser = bluetooth_chooser_new (NULL);
+	g_object_set (adapter->chooser,
+		      "show-searching", FALSE,
+		      "show-device-type", FALSE,
+		      "show-device-category", FALSE,
+		      "show-pairing", TRUE,
+		      "show-connected", TRUE,
+		      "device-category-filter", BLUETOOTH_CATEGORY_PAIRED_OR_TRUSTED,
+		      NULL);
 
-	model = bluetooth_client_get_device_filter_model(client,
-				adapter->proxy, device_filter, NULL, NULL);
-	g_signal_connect(G_OBJECT(model), "row-changed",
-				G_CALLBACK(row_callback), adapter);
-	tree = create_tree(model, FALSE);
-	g_object_unref(model);
+	g_signal_connect (adapter->chooser, "notify::device-selected",
+			  G_CALLBACK(device_selected_cb), adapter);
 
-	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
-	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
+	gtk_table_attach(GTK_TABLE(table), adapter->chooser, 0, 1, 1, 2,
+			 GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
 
-	g_signal_connect(G_OBJECT(selection), "changed",
-				G_CALLBACK(select_callback), adapter);
-
-	adapter->selection = selection;
 	adapter->devices_table = table;
 
-	gtk_container_add(GTK_CONTAINER(scrolled), tree);
-
 	buttonbox = gtk_vbutton_box_new();
 	gtk_button_box_set_layout(GTK_BUTTON_BOX(buttonbox),
 						GTK_BUTTONBOX_START);
@@ -565,20 +456,6 @@ static void create_adapter(adapter_data *adapter)
 
 	adapter->button_disconnect = button;
 
-	button = gtk_button_new_with_label(_("Trust"));
-	image = gtk_image_new_from_stock(GTK_STOCK_ABOUT,
-						GTK_ICON_SIZE_BUTTON);
-	gtk_button_set_image(GTK_BUTTON(button), image);
-	gtk_box_pack_end(GTK_BOX(buttonbox), button, FALSE, FALSE, 0);
-	gtk_button_box_set_child_secondary(GTK_BUTTON_BOX(buttonbox),
-								button, TRUE);
-	gtk_widget_set_sensitive(button, FALSE);
-
-	g_signal_connect(G_OBJECT(button), "clicked",
-				G_CALLBACK(trusted_callback), adapter);
-
-	adapter->button_trusted = button;
-
 	button = gtk_button_new_with_label(_("Delete"));
 	image = gtk_image_new_from_stock(GTK_STOCK_DELETE,
 						GTK_ICON_SIZE_BUTTON);
@@ -599,10 +476,8 @@ static void create_adapter(adapter_data *adapter)
 
 	gtk_widget_show_all(mainbox);
 
-	if (adapter->bring_forward != FALSE) {
+	if (adapter->is_default != FALSE)
 		gtk_notebook_set_current_page (GTK_NOTEBOOK (adapter->notebook), page_num);
-		adapter->bring_forward = FALSE;
-	}
 }
 
 static void update_visibility(adapter_data *adapter)
@@ -626,7 +501,6 @@ static void update_visibility(adapter_data *adapter)
 
 	gtk_toggle_button_set_inconsistent (GTK_TOGGLE_BUTTON (adapter->button_discoverable), inconsistent);
 	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (adapter->button_discoverable), enabled);
-	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (adapter->button_powered), adapter->powered);
 
 	unblock_signals(adapter);
 }
@@ -636,17 +510,7 @@ static void property_changed(DBusGProxy *proxy, const char *property,
 {
 	adapter_data *adapter = user_data;
 
-	if (g_str_equal(property, "Name") == TRUE) {
-		const gchar *name = g_value_get_string(value);
-
-		update_tab_label(GTK_NOTEBOOK(adapter->notebook),
-							adapter->child, name);
-
-		gtk_entry_set_text(GTK_ENTRY(adapter->entry),
-							name ? name : "");
-
-		adapter->name_changed = 0;
-	} else if (g_str_equal(property, "Powered") == TRUE) {
+	if (g_str_equal(property, "Powered") == TRUE) {
 		gboolean powered = g_value_get_boolean(value);
 
 		adapter->powered = powered;
@@ -671,37 +535,21 @@ static void property_changed(DBusGProxy *proxy, const char *property,
 	}
 }
 
-static void
-add_killswitch (GtkNotebook *notebook)
-{
-	if (killswitch_page == NULL)
-		killswitch_page = create_killswitch_page ();
-	if (killswitch_page && gtk_tree_model_iter_n_children (adapter_model, NULL) == 0) {
-		GtkWidget *label;
-
-		label = gtk_label_new (_("Bluetooth Status"));
-		gtk_notebook_prepend_page(notebook, killswitch_page, label);
-		has_killswitch_page = TRUE;
-	}
-}
-
 static adapter_data *adapter_alloc(GtkTreeModel *model,
 		GtkTreePath *path, GtkTreeIter *iter, gpointer user_data)
 {
 	DBusGProxy *proxy;
 	adapter_data *adapter;
 
-	adapter = g_new0(adapter_data, 1);
+	gtk_tree_model_get(model, iter,
+			   BLUETOOTH_COLUMN_PROXY, &proxy,
+			   -1);
 
-	gtk_tree_model_get(model, iter, BLUETOOTH_COLUMN_PROXY, &proxy, -1);
-
-	if (proxy == NULL) {
-		g_free(adapter);
+	if (proxy == NULL)
 		return NULL;
-	}
 
+	adapter = g_new0(adapter_data, 1);
 	adapter->notebook = user_data;
-
 	adapter->reference = gtk_tree_row_reference_new(model, path);
 	adapter->proxy = proxy;
 
@@ -738,20 +586,11 @@ static void adapter_added(GtkTreeModel *model, GtkTreePath *path,
 					GtkTreeIter *iter, gpointer user_data)
 {
 	adapter_data *adapter;
-	int num;
 
 	adapter = adapter_alloc(model, path, iter, user_data);
 	if (adapter == NULL)
 		return;
 
-	num = gtk_notebook_page_num (GTK_NOTEBOOK (user_data), killswitch_page);
-	if (num != -1) {
-		gtk_notebook_remove_page (GTK_NOTEBOOK (user_data), num);
-		adapter->bring_forward = TRUE;
-		has_killswitch_page = FALSE;
-		killswitch_page = NULL;
-	}
-
 	/* XXX This is needed so that we can run dbus_g_proxy_add_signal()
 	 * for "PropertyChanged" on the adapter, remove when we have some
 	 * decent D-Bus bindings */
@@ -783,7 +622,9 @@ static void adapter_removed(GtkTreeModel *model, GtkTreePath *path,
 		adapter->reference = NULL;
 
 		gtk_notebook_remove_page(notebook, i);
-		add_killswitch (notebook);
+		set_current_page (notebook);
+		/* When the default adapter changes, the adapter_changed signal will
+		 * take care of setting the right page */
 
 		g_signal_handlers_disconnect_by_func(adapter->proxy,
 						property_changed, adapter);
@@ -793,60 +634,171 @@ static void adapter_removed(GtkTreeModel *model, GtkTreePath *path,
 	}
 }
 
+static void
+adapter_changed (GtkTreeModel *model,
+		 GtkTreePath  *path,
+		 GtkTreeIter  *iter,
+		 GtkNotebook  *notebook)
+{
+	DBusGProxy *proxy;
+	int i, count;
+	gboolean is_default, powered;
+	char *name;
+
+	count = gtk_notebook_get_n_pages(notebook);
+
+	gtk_tree_model_get(model, iter,
+			   BLUETOOTH_COLUMN_PROXY, &proxy,
+			   BLUETOOTH_COLUMN_DEFAULT, &is_default,
+			   BLUETOOTH_COLUMN_POWERED, &powered,
+			   BLUETOOTH_COLUMN_NAME, &name,
+			   -1);
+
+	for (i = 0; i < count; i++) {
+		GtkWidget *widget;
+		adapter_data *adapter;
+
+		widget = gtk_notebook_get_nth_page(notebook, i);
+		if (widget == NULL)
+			continue;
+
+		adapter = g_object_get_data(G_OBJECT(widget), "adapter");
+		if (adapter == NULL)
+			continue;
+
+		adapter->is_default = is_default;
+
+		if (proxy == NULL || adapter->proxy == NULL)
+			continue;
+
+		if (g_str_equal (dbus_g_proxy_get_path (proxy), dbus_g_proxy_get_path (adapter->proxy)) != FALSE) {
+			if (is_default != FALSE && powered != FALSE)
+				gtk_notebook_set_current_page (notebook, i);
+			/* We usually get an adapter_added before the device
+			 * is powered, so we set the name here instead */
+			if (name)
+				gtk_entry_set_text(GTK_ENTRY(adapter->entry), name);
+		}
+	}
+
+	g_object_unref (proxy);
+	g_free (name);
+}
+
 static void
 button_clicked_cb (GtkButton *button, gpointer user_data)
 {
 	gtk_widget_set_sensitive (GTK_WIDGET (user_data), FALSE);
-	bluetooth_killswitch_set_state (killswitch, KILLSWITCH_STATE_NOT_KILLED);
+	bluetooth_killswitch_set_state (killswitch, KILLSWITCH_STATE_UNBLOCKED);
 }
 
-static GtkWidget *
-create_killswitch_page (void)
+static gboolean
+set_sensitive_now (gpointer user_data)
+{
+	gtk_widget_set_sensitive (GTK_WIDGET (user_data), TRUE);
+	return FALSE;
+}
+
+static void
+killswitch_state_changed (BluetoothKillswitch *killswitch,
+			  KillswitchState state,
+			  gpointer user_data)
+{
+	if (state != KILLSWITCH_STATE_UNBLOCKED)
+		g_timeout_add_seconds (3, set_sensitive_now, user_data);
+}
+
+static void
+create_killswitch_page (GtkNotebook *notebook)
 {
 	GtkWidget *mainbox;
 	GtkWidget *vbox;
 	GtkWidget *label;
 	GtkWidget *button;
 
-	killswitch = bluetooth_killswitch_new ();
-	if (bluetooth_killswitch_has_killswitches (killswitch) == FALSE) {
-		/* No killswitches */
-		g_object_unref (killswitch);
-		return NULL;
+	GtkBuilder *xml;
+
+	xml = gtk_builder_new ();
+	if (gtk_builder_add_from_file (xml, "properties-adapter-off.ui", NULL) == 0) {
+		if (gtk_builder_add_from_file (xml, PKGDATADIR "/properties-adapter-off.ui", NULL) == 0) {
+			g_warning ("Failed to load properties-adapter-off.ui");
+			return;
+		}
 	}
 
+	vbox = GTK_WIDGET (gtk_builder_get_object (xml, "table1"));
+
 	mainbox = gtk_vbox_new(FALSE, 24);
 	gtk_container_set_border_width(GTK_CONTAINER(mainbox), 12);
 
-	vbox = gtk_vbox_new(FALSE, 6);
-	gtk_box_pack_start(GTK_BOX(mainbox), vbox, FALSE, FALSE, 0);
-
 	label = create_label(_("Bluetooth is disabled"));
-	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
+	gtk_box_pack_start(GTK_BOX(mainbox), label, FALSE, FALSE, 0);
 
-	button = gtk_button_new_with_label ("Turn Bluetooth on");
-	g_signal_connect (G_OBJECT (button), "clicked",
+	gtk_box_pack_start(GTK_BOX(mainbox), vbox, TRUE, TRUE, 0);
+
+	button = GTK_WIDGET (gtk_builder_get_object (xml, "button1"));
+	g_signal_connect (button, "clicked",
 			  G_CALLBACK (button_clicked_cb), button);
-	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
+	g_signal_connect (killswitch, "state-changed",
+			  G_CALLBACK (killswitch_state_changed), button);
 
-	return mainbox;
+	gtk_widget_show_all (mainbox);
+
+	gtk_notebook_append_page(notebook, mainbox, NULL);
+}
+
+static void
+create_no_adapter_page (GtkNotebook *notebook)
+{
+	GtkWidget *mainbox;
+	GtkWidget *vbox;
+	GtkWidget *label;
+	GtkBuilder *xml;
+
+	xml = gtk_builder_new ();
+	if (gtk_builder_add_from_file (xml, "properties-no-adapter.ui", NULL) == 0) {
+		if (gtk_builder_add_from_file (xml, PKGDATADIR "/properties-no-adapter.ui", NULL) == 0) {
+			g_warning ("Failed to load properties-no-adapter.ui");
+			return;
+		}
+	}
+	vbox = GTK_WIDGET (gtk_builder_get_object (xml, "table1"));
+
+	mainbox = gtk_vbox_new(FALSE, 24);
+	gtk_container_set_border_width(GTK_CONTAINER(mainbox), 12);
+
+	label = create_label(_("No Bluetooth adapters present"));
+	gtk_box_pack_start(GTK_BOX(mainbox), label, FALSE, FALSE, 0);
+
+	gtk_box_pack_start(GTK_BOX(mainbox), vbox, TRUE, TRUE, 0);
+
+	gtk_widget_show_all (mainbox);
+
+	gtk_notebook_append_page(notebook, mainbox, NULL);
 }
 
 void setup_adapter(GtkNotebook *notebook)
 {
+	killswitch = bluetooth_killswitch_new ();
+
+	/* Create our static pages first */
+	create_killswitch_page (notebook);
+	create_no_adapter_page (notebook);
+
 	client = bluetooth_client_new();
 
 	adapter_model = bluetooth_client_get_adapter_model(client);
 
 	g_signal_connect_after(G_OBJECT(adapter_model), "row-inserted",
 					G_CALLBACK(adapter_added), notebook);
-
 	g_signal_connect_after(G_OBJECT(adapter_model), "row-deleted",
 					G_CALLBACK(adapter_removed), notebook);
+	g_signal_connect_after (G_OBJECT(adapter_model), "row-changed",
+				G_CALLBACK (adapter_changed), notebook);
 
 	gtk_tree_model_foreach(adapter_model, adapter_insert, notebook);
 
-	add_killswitch (notebook);
+	set_current_page (notebook);
 }
 
 void cleanup_adapter(void)
@@ -855,8 +807,6 @@ void cleanup_adapter(void)
 
 	g_object_unref(client);
 
-	if (has_killswitch_page == FALSE && killswitch_page != NULL)
-		gtk_widget_destroy (killswitch_page);
-	killswitch_page = NULL;
+	g_object_unref (killswitch);
 }
 
